from stringology.align import align, mismatches


class SpSim:
    '''SpSim is a spelling similarity measure for identifying cognates by
    learning cross-language spelling differences.

    >>> sim = SpSim()
    >>> sim('phase', 'fase')
    0.6

    Learning contextualized spelling differences from an example:

    >>> sim.learn([('alpha', 'alfa')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', 'la')]

    SpSim has learned that 'ph' may be replaced by 'f' if 'ph' comes after 'l'
    and before 'a'.

    Generalizing the admissible contexts of known spelling differences:

    >>> sim.learn([('phase', 'fase')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', '*a')]

    SpSim has learned that 'ph' may be replaced by 'f' if it comes before 'a'.

    >>> sim.learn([('photo', 'foto')])
    >>> sorted(sim.diffs.items())
    [('ph\\tf', '**')]

    SpSim has learned that 'ph' may be replaced by 'f'.

    >>> sim('phenomenal', 'fenomenal')
    1.0

    '''

    def __init__(self, examples=None):
        self.diffs = {}
        if examples:
            self.learn(examples)

    def __call__(self, a, b):
        d = 0  # total distance
        for n, diff, ctxt in SpSim._get_diffs(a, b):
            if not SpSim._match_context(ctxt, self.diffs.get(diff, None)):
                d += n
        return 1.0 - d / max(1, len(a), len(b))

    def learn(self, examples):
        for a, b in examples:
            for n, diff, ctxt in SpSim._get_diffs(a, b):
                learned = self.diffs.get(diff, None)
                self.diffs[diff] = SpSim._generalize_context(learned, ctxt)

    def _get_diffs(a, b):
        alignment = align('^' + a + '$', '^' + b + '$', gap=' ')
        for mma, mmb in mismatches(*alignment, context=1):
            n = len(mma) - 2  # discount the left and right context chars
            diff = (mma[1:-1] + '\t' + mmb[1:-1]).replace(' ', '')
            ctxt = mma[0] + mma[-1]
            yield n, diff, ctxt

    def _match_context(ctxt, learned):
        return (
            learned and
            learned[0] in '*' + ctxt[0] and
            learned[1] in '*' + ctxt[1]
        )

    def _generalize_context(learned, ctxt):
        if not learned:
            return ctxt
        else:
            lft = '*' if learned[0] != ctxt[0] else learned[0]
            rgt = '*' if learned[1] != ctxt[1] else learned[1]
            return lft + rgt




import munkres
from spsim.spsim import SpSim


class PhraseSpSim(SpSim):
    '''
    Measures the similarity between pairs of phrases using SpSim for comparing
    words.

    A phrase is a sequence one or more words.

                January 2010, Luís Gomes <luismsgomes@gmail.com>

    Works as follows:
    1 - Compute the distance between each word A of one phrase and each word B
        of the  other phrase using (1 - SpSim(A, B)) * max(len(A), len(B)).
    2 - Compute the one-to-one assignment of words from one phrase to words
        of the other phrase that minimizes the sum of all pair distances. If
        the phrases have a different number of words, then we add null words
        to the shortest phrase. The distance between any word C and null is
        len(C).
    3 - The similarity score is computed as 1 - D/L, where D is the minimum
        total distance (computed in 2) and L is the length of the longest
        phrase, excluding whitespace.

    Step 2 is computed using the Munkres algorithm (aka Hungarian). Note that
    the problem solved in step 2 is widely known as "the assignment problem".

    >>> sim = PhraseSpSim()
    >>> sim.learn([('photo', 'foto'), ('alpha', 'alfa'), ('pangea', 'pangeia')])
    >>> sorted(sim.diffs.items())
    [('\\ti', 'ea'), ('ph\\tf', '**')]

    SpSim has learned that 'ph' may be replaced by 'f' and 'ea' by 'eia'.

    >>> sim('phenomenal idea', 'ideia fenomenal')
    1.0

    '''
    def __init__(self, examples=None):
        super().__init__(examples=examples)
        self._dist_cache = {}

    def __call__(self, a, b):
        if isinstance(a, str):
            a, b = a.split(), b.split()
        # compute the length of the longest phrase (excluding spaces)
        L = max(sum(map(len, a)), sum(map(len, b)))
        if 1 == len(a) == len(b): return 1.0 - self._dist(a[0], b[0]) / L
        # let's obtain the (square) distance matrix
        if len(a) > len(b): b.extend(['']*(len(a)-len(b)))
        elif len(a) < len(b): a.extend(['']*(len(b)-len(a)))
        M = [[self._dist(p, q) for q in b] for p in a]
        # now compute the one to one assignment between a and b that
        # minimizes the sum of distances (this is "the assignment problem")
        D = sum(M[row][col] for row, col in munkres.Munkres().compute(M))
        # sometimes we might have a distance that is greater than the length
        # of either expressions if we happen to have a compound word in one
        # side; therefore we must limit D to be at most equal to L;
        # to mitigate this problem we should split compounds beforehand
        if D > L: D = L
        return 1.0 - D / L

    def _dist(self, a, b):
        # first check if this is a obvious case (those aren't cached)
        if a == b: return 0  # pragma: no cover
        if '' == a: return len(b)  # pragma: no cover
        if '' == b: return len(a)  # pragma: no cover

        key = a + '\t' + b
        d = self._dist_cache.get(key, None)
        if d is None:
            d = (1.0 - super().__call__(a, b)) * max(len(a), len(b))
            self._dist_cache[key] = d
        return d

    def learn(self, examples):
        super().learn(examples)
        self._dist_cache = {}



"""
Usage: spsim <examples-file> [<input-file> [<output-file>]]
"""

import docopt
import openfile
import logging
import sys
from spsim2 import PhraseSpSim


logger = logging.getLogger("spsim")


def main(argv=sys.argv):
    logging.basicConfig()
    opts = docopt.docopt(__doc__)
    sim = PhraseSpSim()
    with (openfile.openfile(opts["<examples-file>"])) as lines:
        sim.learn(
            line.rstrip("\n").split("\t")[:2] for line in lines if "\t" in line
        )
    inputfile = openfile.openfile(opts["<input-file>"])
    outputfile = openfile.openfile(opts["<output-file>"], "wt")
    with inputfile, outputfile:
        for num, line in enumerate(inputfile, start=1):
            cols = line.rstrip('\n').split('\t')
            if 2 > len(cols):
                logger.warning(
                    'line %d of %s has less than 2 columns; skipping',
                    num, opts["<input-file>"]
                )
                continue
            cols.append(sim(cols[0], cols[1]))
            print(*cols, sep='\t', file=outputfile)


if __name__ == '__main__':
    main()
